#ifndef env_hpp
#define env_hpp

#include <vector>
#include <utility>
#include <set>
#include <map>

#include "point.hpp"
#include "team.hpp"

/* Probability of not randomly selecting a specific exemplar. */
#define PFAIL 1e-31

class env
{
protected:
  
  /* Act given the (starting) point, action, and (current) state. Return
     rewards and write whether or not end state has been reached. */
  virtual double act(point *, long, vector < double > &, bool &) = 0;  
  
public:
  
  env() 
  {};
  virtual ~env() 
  {};
  
  /* Select a point in the environment with uniform probability. */
  virtual point * initUniformPoint(long, set < long > &) = 0;  
  
  double evaluate(team *, point *);
  
  /* Test a team on the point in the environment. */
  virtual double test(team *, string) = 0;
};

class explicitEnv : public env
{
protected:
  
  string _setName;
  int _dim;

  /* The set of points defining the environment. The point id should be the 
     index of the point in this vector (and the underlying dataset). */
  vector < pair < double *, long > * > _pointSet;

  /* Record index of points in _pointSet grouped by label. The dimension
     of _index should be the number of points. The dimension of *_index[i]
     should be the number of points of label _labels[i]. This is done to
     speed up sampling of minority classes. */
  vector < vector < long > * > _index;
  
  /* Labels sorted into ascending order. */
  vector < long > _labels;
  /* Count the number of each label at each index (order matching _labels) */
  vector < long > _labelCount;
  /* Map label->index to find a label in _labels and _labelCount. */
  map < long, long > _labelMap;

  long _maxTries;
  
public:

  explicitEnv(int dim, string setName) 
    : _setName(setName), _dim(dim), _maxTries(0) {};
  
  inline long numLabels() 
  { return _labels.size(); }
  
  /* Get label at some index in _labels. */
  inline long getLabel(long i) 
  { return _labels[i]; }
  /* Get index in _labels and _labelCount of some label. */
  long getLabelIndex(long);
  /* Get the number of points of a given label. */
  inline long getLabelCount(long label) 
  { return _labelCount[getLabelIndex(label)]; }
  
  inline long size() 
  { return _pointSet.size(); }
  
  point * initUniformPoint(long, set < long > &);
  point * initUniformPointLabel(long, set < long > &, long);
};

class datasetEnv : public explicitEnv
{
public:
  
  datasetEnv(string, int, long);
  ~datasetEnv();
  
  double act(point *, long, vector <double > &, bool &);

  double test(team *, string);
};

class implicitEnv : public env
{ 
public:
  
};

#endif
