#include "learner.hpp"

/* The source register is taken as srcVal % REGISTERS so the max value
   of srcVal should be one less a multiple of REGISTERS so that there
   is no bias towards any register.  For example, if REGISTERS is 4
   and there srcVal can take values 0 to 4, then register 0 is
   expected to occur more frequently. The destination register is
   simply the bits in the dst field converted to ulong. */

/* 0 0000 0000 0001 */
const instruction learner::_modeMask(0x1);   
/* 0 0000 0000 1110 */
const instruction learner::_opMask(0xE);     
/* 0 0000 0111 0000 */
const instruction learner::_dstMask(0x70);   
/* 1 1111 1000 0000 */
const instruction learner::_srcMask(0x1F80); 

/* Rx <- op Rx Ry */
const instruction learner::_mode0(0x0); 
/* Rx <- op Rx Iy */
const instruction learner::_mode1(0x1); 

const instruction learner::_opSum(0x0);
const instruction learner::_opDiff(0x1);
const instruction learner::_opProd(0x2);
const instruction learner::_opDiv(0x3);
const instruction learner::_opMod(0x4);
const instruction learner::_opCos(0x5);
const instruction learner::_opExp(0x6);
const instruction learner::_opLog(0x7);

const short learner::_modeShift = 0;
const short learner::_opShift = learner::_modeMask.count();
const short learner::_dstShift = learner::_modeMask.count() + learner::_opMask.count();
const short learner::_srcShift = learner::_modeMask.count() + learner::_opMask.count() + learner::_dstMask.count();

long learner::_count = 0;

learner::learner(long gtime, 
		 long action,
		 int maxProgSize)
  : _id(_count++), _gtime(gtime), _action(action), _nrefs(0)
{  
  instruction *in;
  
  int progSize = 1 + ((int) (drand48() * maxProgSize));
  
  for(int i = 0; i < progSize; i++)
    {
      in = new instruction();
      
      for(int j = 0; j < in->size(); j++)
	if(drand48() < 0.5) in->flip(j);
      
      _bid.push_back(in);
    }
}

learner::learner(long gtime,
		 learner &plr)
  : _id(_count++), _gtime(gtime), _action(plr.action()), _nrefs(0)
{  
  vector < instruction * > :: iterator initer, initerend;
  
  for(initer = plr._bid.begin(), initerend = plr._bid.end(); 
      initer != initerend; initer++)
    _bid.push_back(new instruction(**initer));
}

bool learner::muBid(double pBidDelete,
		    double pBidAdd,
		    double pBidSwap,
		    double pBidMutate,
		    int maxProgSize)
{
  bool changed = false;
  
#ifdef MYDEBUG
  cout << "learner:muBid before";
  for(int i = 0; i < _bid.size(); i++)
    cout << " " << i << ":" << _bid[i] << ":" << *_bid[i];
  cout << endl;
#endif

  /* Remove random instruction. */
  if(_bid.size() > 1 && drand48() < pBidDelete)
    {
      int i = (int) (drand48() * _bid.size());

      delete *(_bid.begin() + i);
      _bid.erase(_bid.begin() + i);

      changed = true;

#ifdef MYDEBUG
      cout << "learner:muBid delete " << i << endl;
#endif
    }
  
  /* Insert random instruction. */
  if(_bid.size() < maxProgSize && drand48() < pBidAdd)
    {
      instruction *instr = new instruction();
      
      for(int j = 0; j < instr->size(); j++)
        if(drand48() < 0.5) instr->flip(j);
      
      int i = (int) (drand48() * (_bid.size() + 1));
      
      _bid.insert(_bid.begin() + i, instr);
      
      changed = true;
      
#ifdef MYDEBUG
      cout << "learner:muBid add " << i << " " << instr << ":" << *instr << endl;
#endif
    }
  
  /* Flip single bit of random instruction. */
  if(drand48() < pBidMutate)
    {
      int i = (int) (drand48() * _bid.size());
      int j = (int) (drand48() * _bid[0]->size());
      
      _bid[i]->flip(j);
      
      changed = true;

#ifdef MYDEBUG
      cout << "learner:muBid mutate " << i << ", " << j << endl;
#endif
    }

  /* Swap positions of two instructions. */
  if(_bid.size() > 1 && drand48() < pBidSwap)
    {
      int i = (int) (drand48() * _bid.size());
      
      int j;
      do
	{
	  j = (int) (drand48() * _bid.size());
	} while(i == j);
      
      instruction *tmp;
      
      tmp = _bid[i];
      _bid[i] = _bid[j];
      _bid[j] = tmp;

      changed = true;

#ifdef MYDEBUG
      cout << "learner:muBid swap " << i << ", " << j << endl;
#endif
    }

#ifdef MYDEBUG
  cout << "learner:muBid after";
  for(int i = 0; i < _bid.size(); i++)
    cout << " " << i << ":" << _bid[i] << ":" << *_bid[i];
  cout << endl;
#endif

  return changed;
}
		 
learner::~learner()
{
  for(int i = 0; i < _bid.size(); i++)
    delete _bid[i];
}

double learner::bid(double *feature, 
		    int dim, 
		    double *REG)
{
  /* Zero registers. */
  memset(REG, 0, REGISTERS * sizeof(double));

  return 1 / (1 + exp(-run(_bid, feature, dim, REG)));
}

double learner::run(vector < instruction * > &prog, 
		    double *feature, 
		    int dim, 
		    double *REG)
{  
  vector < instruction * > :: iterator initer, initerend;
  
  instruction mode;
  instruction op;
  
  int dstReg;
  double srcVal;
  
#ifdef SHOWEXEC
  cout << *this << endl << "FEATURE";
  for(int i = 0; i < dim; i++)
    cout << " " << feature[i];
  cout << endl << "REG";
  for(int i = 0; i < REGISTERS; i++)
    cout << " " << REG[i];
  cout << endl;
#endif
  
  for(initer = prog.begin(), initerend = prog.end(); 
      initer != initerend; initer++)
    {
      mode = (**initer & _modeMask) >> _modeShift; 
      op = (**initer & _opMask) >> _opShift;

      /* Should be between 0 and REGISTERS - 1. */

      dstReg = ((**initer & _dstMask) >> _dstShift).to_ulong();

#ifdef SHOWEXEC
      cout << **initer << " R[" << dstReg << "] <- R[" << dstReg << "] ";
#endif

      if(mode == _mode0) // Rx <- op Rx Ry
	{
	  srcVal = REG[((**initer & _srcMask) >> _srcShift).to_ulong() % REGISTERS];
#ifdef SHOWEXEC
	  cout << "R[" << ((**initer & _srcMask) >> _srcShift).to_ulong() % REGISTERS << "] ";
#endif
	}
      else // Rx <- op Rx Iy
	{
	  srcVal = feature[((**initer & _srcMask) >> _srcShift).to_ulong() % dim];
#ifdef SHOWEXEC
	  cout << "I[" <<  ((**initer & _srcMask) >> _srcShift).to_ulong() % dim << "] ";
#endif
	}
      
      if(op == _opSum)
	{
	  REG[dstReg] = REG[dstReg] + srcVal;
#ifdef SHOWEXEC
	  cout << "sum ";
#endif
	}
      else if(op == _opDiff)
	{
	  REG[dstReg] = REG[dstReg] - srcVal;
#ifdef SHOWEXEC
	  cout << "diff ";
#endif
	}
      else if(op == _opProd)
	{
	  REG[dstReg] = REG[dstReg] * srcVal;
#ifdef SHOWEXEC
	  cout << "prod ";
#endif
	}
      else if(op == _opDiv)
	{
	  REG[dstReg] = REG[dstReg] / srcVal;
#ifdef SHOWEXEC
	  cout << "div ";
#endif
	}
      else if(op == _opMod)
	{
	  REG[dstReg] = fmod(REG[dstReg], srcVal);
#ifdef SHOWEXEC
	  cout << "mod ";
#endif
	}
      else if(op == _opCos)
	{
	  REG[dstReg] = cos(srcVal);
#ifdef SHOWEXEC
	  cout << "cos ";
#endif
	}
      else if(op == _opExp)
	{
	  REG[dstReg] = exp(srcVal);
#ifdef SHOWEXEC
	  cout << "exp ";
#endif
	}
      else if(op == _opLog)
	{
	  REG[dstReg] = log(fabs(srcVal));
#ifdef SHOWEXEC
	  cout << "log ";
#endif
	}
      else
	{
	  die(__FILE__, __FUNCTION__, __LINE__, "bad operation");
	}
      
      if(isfinite(REG[dstReg]) == 0)
	REG[dstReg] = 0;
      
#ifdef SHOWEXEC
      cout << "REG";
      for(int i = 0; i < REGISTERS; i++)
	cout << " " << REG[i];
      cout << endl;
#endif
    }
  
  return REG[0];
}

ostream & operator<<(ostream &os, 
		     const learner &lr)
{
  os << "[" << lr._id << " " << lr._action << "]";

  return os;
}

