#ifndef scm_hpp
#define scm_hpp

#include <map>
#include <set>
#include <algorithm>

#include "point.hpp"
#include "team.hpp"
#include "learner.hpp"
#include "env.hpp"

class scm
{
  bool genTeams(long, team *);
  void getDist(map < point *, vector < short > * > &);
  template < class ptype, class vtype > void findParetoFront(map < ptype *, vector < vtype > * > &, set < ptype * > &, set < ptype * > &, double);
  template < class vtype > bool dominated(vector < vtype > &, vector < vtype > &, bool &, double);
  template < class ptype, class vtype > void sharingScore(map < ptype *, vector < vtype > * > &, map < ptype *, double > &, set < ptype * > &, set < ptype * > &);
  template < class ptype > void selPareto(set < ptype * > &, map < ptype *, double > &, int, set < ptype * > &);
  void cleanup();
  void finalize();

protected:

  set < point * > _P;
  set < team * > _M;
  set < learner * > _L;
  
  /* Point population size. */
  int _Psize;
  /* Team population size. */
  int _Msize;
  /* Probability of learner deletion. */
  double _pd;
  /* Probability of learner addition. */
  double _pa;
  /* Probability of action mutation. */
  double _mua;
  /* Maximum team size. */
  int _omega;
  /* Number of training epochs. */
  long _t;
  /* Point generation gap. */
  int _Pgap;
  /* Team generation gap. */
  int _Mgap;

  /* GP parameters. */
  int _maxProgSize;
  double _pBidDelete;
  double _pBidAdd;
  double _pBidSwap;
  double _pBidMutate;

  long _statMod;

  int _seed;

  /* Some stats. */

  long _mdom;
  long _pdom;

  int _pointFrontSize;
  int _teamFrontSize;

  virtual void initPoints() = 0;
  void initTeams();
  virtual void genPoints(long) = 0;
  void genTeams(long);
  virtual void evaluate(team *, point *) = 0;
  void selPoints(long);
  void selTeams(long);
  virtual void test() = 0;
  virtual void stats(long) = 0;
  virtual long numActions() = 0;
  
public:
  
  scm(map < string, string > &);
  virtual ~scm();
  
  void go();
};

class scmImplicit : public scm
{
public:
  scmImplicit();
};

class scmExplicit : public scm
{
protected:

  /* At this level actions 0, 1, ..., _numActions - 1 are assumed. */
  long _numActions;
  
  explicitEnv *_trainEnv;
  explicitEnv *_testEnv;
  
public:
  
  scmExplicit(map < string, string > &);
  ~scmExplicit();

  inline long numActions()
  { return _numActions; }
  
  void initPoints();
  void genPoints(long);
  inline void evaluate(team *m, point *p) 
  { m->setOutcome(p, _trainEnv->evaluate(m, p)); }
  void test();
  void stats(long);
};

template < class ptype > struct lessThan : public binary_function < ptype *, ptype *, bool >
{
  bool operator()(ptype *lhs, ptype *rhs) { return lhs->key() < rhs->key(); }
};

#endif
